import numpy as np
import pytest
import scipy.sparse as sp

import kaczmarz

np.random.seed(0)

strategies = []
for name in dir(kaczmarz):
    if name.startswith("_"):
        continue
    attr = getattr(kaczmarz, name)
    if attr == kaczmarz.Base:
        continue
    if not issubclass(attr, kaczmarz.Base):
        continue
    strategies.append(attr)


def orthogonal_rows():
    examples = []

    A = [
        [1, 0, 0],
        [0, 1, 0],
    ]
    x = [1, 2, 0]
    examples.append((A, x))
    A = [
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1],
    ]
    x = [1, 2, 3]
    examples.append((A, x))
    A = [
        [1, 1, 0],
        [-1, 1, 0],
    ]
    x = [1, 2, 0]
    examples.append((A, x))
    A = [
        [1, 1, 0],
        [-1, 1, 0],
        [0, 0, 1],
    ]
    x = [1, 2, 3]
    examples.append((A, x))

    return examples


def underdetermined():
    examples = []

    A = [
        [0, 0, 1, 1],
    ]
    x = [0, 0, 1, 1]
    examples.append((A, x))
    A = [
        [0, 0, 0, 2],
    ]
    x = [0, 0, 0, 0.5]
    examples.append((A, x))
    A = [
        [0, 0, 0, 2],
        [1, 0, 0, 1],
    ]
    x = [1, 0, 0, 0.5]
    examples.append((A, x))
    A = [
        [2, 1, 2, 1],
        [1, 2, 1, 2],
        [2, 1, 2, 1],
    ]
    x = [0.5, 0.25, 0.5, 0.25]
    examples.append((A, x))

    return examples


def overdetermined():
    examples = []
    A = [
        [2],
        [1],
        [2],
    ]
    x = [1]
    examples.append((A, x))
    A = [
        [2, 1],
        [1, 2],
        [2, 3],
    ]
    x = [2, 3]
    examples.append((A, x))
    return examples


systems = orthogonal_rows() + underdetermined() + overdetermined()


@pytest.mark.parametrize("A,x_exact", systems)
@pytest.mark.parametrize("Strategy", strategies)
def test_solve(A, x_exact, Strategy, allclose):
    """Solvers should accept list-of-lists, np.ndarray, and csr_matrix."""
    row_norms = np.linalg.norm(A, axis=1)
    tol = 1e-5
    Anp = np.array(A)
    Asp = sp.csr_matrix(A)
    b = Anp @ x_exact
    x_approx = Strategy.solve(A, b, tol=tol)
    assert np.linalg.norm(Anp @ (x_approx - x_exact) / row_norms) < tol
    assert allclose(x_approx, x_exact, rtol=10 * tol)
    x_approx = Strategy.solve(Anp, b, tol=tol)
    assert np.linalg.norm(Anp @ (x_approx - x_exact) / row_norms) < tol
    assert allclose(x_approx, x_exact, rtol=10 * tol)
    x_approx = Strategy.solve(Asp, b, tol=tol)
    assert np.linalg.norm(Anp @ (x_approx - x_exact) / row_norms) < tol
    assert allclose(x_approx, x_exact, rtol=10 * tol)


@pytest.mark.parametrize("A,x_exact", systems)
@pytest.mark.parametrize("Strategy", strategies)
def test_iterates_converge_monotonically(A, x_exact, Strategy):
    """Check that solver works on list-of-lists, np.ndarray, and csr_matrix."""
    Anp = np.array(A)
    b = Anp @ x_exact
    errors = [np.linalg.norm(xk - x_exact) for xk in Strategy.iterates(Anp, b)]
    assert errors[1:] <= errors[:-1]



"""A module providing selection strategies for the Kaczmarz algorithm."""

from collections import deque

import numpy as np

import kaczmarz


class Cyclic(kaczmarz.Base):
    """Cycle through the equations of the system in order, repeatedly.

    References
    ----------
    1. S. Kaczmarz.
       "Angenäherte Auflösung von Systemen linearer Gleichungen."
       *Bulletin International de l’Académie Polonaise
       des Sciences et des Lettres.
       Classe des Sciences Mathématiques et Naturelles.
       Série A, Sciences Mathématiques*, 35, 335–357, 1937
    """

    def __init__(self, *base_args, **base_kwargs):
        super().__init__(*base_args, **base_kwargs)
        self._row_index = -1

    def _select_row_index(self, xk):
        self._row_index = (1 + self._row_index) % self._n_rows
        return self._row_index


class MaxDistance(kaczmarz.Base):
    """Choose equations which leads to the most progress.

    This selection strategy is also known as `Motzkin's method`.

    References
    ----------
    1. T. S. Motzkin and I. J. Schoenberg.
       "The relaxation method for linear inequalities."
       *Canadian Journal of Mathematics*, 6:393–404, 1954.
    """

    def _select_row_index(self, xk):
        # TODO: use auxiliary update for the residual.
        residual = self._b - self._A @ self._xk
        return np.argmax(np.abs(residual))


class Random(kaczmarz.Base):
    """Sample equations according to a `fixed` probability distribution.

    Parameters
    ----------
    p : (m,) array_like, optional
        Sampling probability for each equation. Uniform by default.
    """

    def __init__(self, *base_args, p=None, **base_kwargs):
        super().__init__(*base_args, **base_kwargs)
        self._p = p  # p=None corresponds to uniform.

    def _select_row_index(self, xk):
        return np.random.choice(self._n_rows, p=self._p)


class SVRandom(Random):
    """Sample equations with probability proportional to the squared row norms.

    References
    ----------
    1. T. Strohmer and R. Vershynin,
       "A Randomized Kaczmarz Algorithm with Exponential Convergence."
       Journal of Fourier Analysis and Applications 15, 262 2009.
    """

    def __init__(self, *base_args, **base_kwargs):
        super().__init__(*base_args, **base_kwargs)
        squared_row_norms = self._row_norms ** 2
        self._p = squared_row_norms / squared_row_norms.sum()


class UniformRandom(Random):
    """Sample equations uniformly at random."""

    # Nothing to do since uniform sampling is the default behavior of Random.


class Quantile(Random):
    """Reject equations whose normalized residual is above a quantile.

    This algorithm is intended for use in solving corrupted systems of equations.
    That is, systems where a subset of the equations are consistent,
    while a minority of the equations are not.
    Such systems are almost always overdetermined.

    Parameters
    ----------
    quantile : float, optional
        Quantile of normalized residual above which to reject.

    References
    ----------
    1. There will be a reference soon. Keep an eye out for that.
    """

    def __init__(self, *args, quantile=1.0, **kwargs):
        super().__init__(*args, **kwargs)
        self._quantile = quantile

    def _distance(self, xk, ik):
        return np.abs(self._b[ik] - self._A[ik] @ xk)

    def _threshold_distances(self, xk):
        return np.abs(self._b - self._A @ xk)

    def _threshold(self, xk):
        distances = self._threshold_distances(xk)

        return np.quantile(distances, self._quantile)

    def _select_row_index(self, xk):
        ik = super()._select_row_index(xk)

        distance = self._distance(xk, ik)
        threshold = self._threshold(xk)

        if distance < threshold or np.isclose(distance, threshold):
            return ik

        return -1  # No projection please


class SampledQuantile(Quantile):
    """Reject equations whose normalized residual is above a quantile of a random subset of residual entries.

    Parameters
    ----------
    n_samples: int, optional
        Number of normalized residual samples used to compute the threshold quantile.

    References
    ----------
    1. There will be a reference soon. Keep an eye out for that.
    """

    def __init__(self, *args, n_samples=None, **kwargs):
        super().__init__(*args, **kwargs)
        if n_samples is None:
            n_samples = self._n_rows
        self._n_samples = n_samples

    def _threshold_distances(self, xk):
        idxs = np.random.choice(self._n_rows, self._n_samples, replace=False)
        return np.abs(self._b[idxs] - self._A[idxs] @ xk)


class WindowedQuantile(Quantile):
    """Reject equations whose normalized residual is above a quantile of the most recent normalized residual values.

    Parameters
    ----------
    window_size : int, optional
        Number of recent normalized residual values used to compute the threshold quantile.

    Note
    ----
    ``WindowedQuantile`` also accepts the parameters of ``Quantile``.

    References
    ----------
    1. There will be a reference soon. Keep an eye out for that.
    """

    def __init__(self, *args, window_size=None, **kwargs):
        super().__init__(*args, **kwargs)
        if window_size is None:
            window_size = self._n_rows
        self._window = deque([], maxlen=window_size)

    def _distance(self, xk, ik):
        distance = super()._distance(xk, ik)
        self._window.append(distance)
        return distance

    def _threshold_distances(self, xk):
        return self._window



"""
Dummy conftest.py for clapsolver.

If you don't know what this is for, just leave it empty.
Read more about conftest.py under:
https://pytest.org/latest/plugins.html

https://stackoverflow.com/questions/34466027/in-pytest-what-is-the-use-of-conftest-py-files
"""

import numpy as np
import pytest
import scipy.sparse as sp


@pytest.fixture
def eye33():
    return np.eye(3)


@pytest.fixture
def eye23():
    return np.eye(2, 3)


@pytest.fixture
def speye23():
    return sp.csr_matrix(np.eye(2, 3))


@pytest.fixture
def zeros3():
    return np.zeros(3)


@pytest.fixture
def ones3():
    return np.ones(3)


@pytest.fixture()
def ones2():
    return np.ones(2)



"""Top-level package for Kaczmarz Algorithms."""

# Do not edit this string manually, always use bumpversion
# Details in CONTRIBUTING.md
__version__ = "0.7.0"

__author__ = "Jacob Moorman"
__email__ = "jacob@moorman.me"

__license__ = "MIT"
__copyright__ = "Copyright (c) 2020, Jacob Moorman"


from ._abc import Base
from ._variants import (
    Cyclic,
    MaxDistance,
    Quantile,
    Random,
    SampledQuantile,
    SVRandom,
    UniformRandom,
    WindowedQuantile,
)



from abc import ABC, abstractmethod

import numpy as np

from ._normalize import normalize_system


class Base(ABC):
    """A base class for the Kaczmarz algorithm.

    This class cannot be instantiated directly.
    Subclasses should implement :meth:`kaczmarz.Base._select_row_index`.
    Subclasses will typically be constructed using
    :meth:`kaczmarz.Base.iterates` or :meth:`kaczmarz.Base.solve`.

    Parameters
    ----------
    A : (m, n) spmatrix or array_like
        The m-by-n matrix of the linear system.
    b : (m,) or (m, 1) array_like
        Right hand side of the linear system.
    x0 : (n,) or (n, 1) array_like, optional
        Starting guess for the solution.
    tol : float, optional
        Tolerance for convergence, ``norm(normalized_residual) <= tol``.
    maxiter : int or float, optional
        Maximum number of iterations.
    callback : function, optional
        User-supplied function to call after each iteration.
        It is called as ``callback(xk)``,
        where xk is the current solution vector.

    Notes
    -----
    There may be additional parameters not listed above
    depending on the selection strategy subclass.
    """

    def __init__(
        self, A, b, x0=None, tol=1e-5, maxiter=None, callback=None,
    ):
        self._A, self._b, self._row_norms = normalize_system(A, b)
        self._n_rows = len(self._b)

        if x0 is None:
            n_cols = self._A.shape[1]
            x0 = np.zeros(n_cols)
            self._iterate_shape = list(np.shape(b))  # [m,] or [m, 1]
            self._iterate_shape[0] = n_cols
        else:
            x0 = np.array(x0, dtype="float64")
            self._iterate_shape = x0.shape

        self._x0 = x0.ravel()
        self._tol = tol

        if maxiter is None:
            maxiter = 2 * np.log(tol) / np.log(1 - 1 / (10 * min(self._A.shape)))

        self._maxiter = maxiter
        if callback is None:

            def callback(xk):
                return None

        self._callback = callback
        self._k = -1
        self._ik = -1
        self._xk = None

    @property
    def ik(self):
        """int: The index of the row used on the most recent iteration.

        Takes the value -1 if a projection was not performed at iteration ``k``."""
        return self._ik

    @property
    def xk(self):
        """(n,) or (n, 1) array: The most recent iterate.

        The shape will match that of ``x0`` if provided, or ``b`` otherwise.
        """
        return self._xk.copy().reshape(*self._iterate_shape)

    @classmethod
    def iterates(cls, *base_args, **base_kwargs):
        """Get the Kaczmarz iterates.

        Note
        ----
        This method takes the same parameters as :class:`kaczmarz.Base`
        or the subclass from which it is called.
        For example, :meth:`kaczmarz.Cyclic.iterates`
        takes the same arguments as :class:`kaczmarz.Cyclic`.

        Parameters
        ----------
        base_args : tuple
            Positional arguments for :class:`kaczmarz.Base` constructor
            or the subclass in use.
        base_kwargs : dict
            Keyword arguments for :class:`kaczmarz.Base` constructor
            or the subclass in use.

        Returns
        -------
        iterates : iterable((n,) or (n, 1) array)
            An iterable of the Kaczmarz iterates.
            The shapes will match that of ``x0`` if provided,
            or ``b`` otherwise.
        """
        return cls(*base_args, **base_kwargs)

    @classmethod
    def solve(cls, *base_args, **base_kwargs):
        """Solve a linear system of equations using the Kaczmarz algorithm.

        Note
        ----
        This method takes the same parameters as :class:`kaczmarz.Base`
        or the subclass from which it is called.
        For example, :meth:`kaczmarz.Cyclic.solve`
        takes the same arguments as :class:`kaczmarz.Cyclic`.

        Parameters
        ----------
        base_args : tuple
            Positional arguments for :class:`kaczmarz.Base` constructor
            or the subclass in use.
        base_kwargs : dict
            Keyword arguments for :class:`kaczmarz.Base` constructor
            or the subclass in use.

        Returns
        -------
        x : (n,) or (n, 1) array
            The solution to the system ``A @ x = b``.
            The shape will match that of ``x0`` if provided,
            or ``b`` otherwise.
        """
        iterates = cls.iterates(*base_args, **base_kwargs)
        for x in iterates:
            pass
        return x

    def __next__(self):
        """Perform an iteration of the Kaczmarz algorithm.

        Returns
        -------
        xk : (n,) or (n, 1) array
            The next iterate of the Kaczmarz algorithm.
            The shape will match that of ``x0`` if provided,
            or ``b`` otherwise.
        """
        if self._k == -1:
            self._k += 1
            self._xk = self._x0
            self._callback(self.xk)
            return self.xk

        if self._stopping_criterion(self._k, self._xk):
            raise StopIteration

        self._k += 1
        self._ik = self._select_row_index(self._xk)
        if self._ik != -1:
            self._xk = self._update_iterate(self._xk, self._ik)

        self._callback(self.xk)

        return self.xk

    def __iter__(self):
        """Iterator for iterates of the Kaczmarz algorithm."""
        return self

    @abstractmethod
    def _select_row_index(self, xk):
        """Select a row to use for the next Kaczmarz update.

        Parameters
        ----------
        xk : (n,) array
            The current Kaczmarz iterate.

        Returns
        -------
        ik : int
            The index of the next row to use.
        """

    def _update_iterate(self, xk, ik):
        """Apply the Kaczmarz update.

        Parameters
        ----------
        xk : (n,) array
            The current iterate of the Kaczmarz algorithm.
        ik : int
            Row index to use for the update.

        Returns
        -------
        xkp1 : (n,) array
            The next iterate
        """
        ai = self._A[ik]
        bi = self._b[ik]
        return xk + (bi - ai @ xk) * ai

    def _stopping_criterion(self, k, xk):
        """Check if the iteration should terminate.

        Parameters
        ----------
        k : int
            The number of iterations that have passed.
        xk : (n,) array
            The current iterate of the Kaczmarz algorithm.

        Returns
        -------
        stop : bool
            True if the iteration should be terminated.
        """
        if self._k >= self._maxiter:
            return True

        residual = self._b - self._A @ self._xk
        residual_norm = np.linalg.norm(residual)

        if residual_norm < self._tol:
            return True

        return False



import numpy as np
import pytest

import kaczmarz

np.random.seed(0)

quantile_strategies = []
for name in dir(kaczmarz):
    if name.startswith("_"):
        continue
    if "Quantile" not in name:
        continue
    attr = getattr(kaczmarz, name)
    if attr == kaczmarz.Base:
        continue
    if not issubclass(attr, kaczmarz.Base):
        continue
    quantile_strategies.append(attr)


def get_systems():
    examples = []

    A = np.arange(1, 11).reshape(-1, 1)  # [[1], [2], ...]
    x = np.array([1])
    examples.append((A, x))

    A = np.random.normal(size=(10, 2))
    x = np.array([1, 1])
    examples.append((A, x))

    return examples


class Sample100(kaczmarz.SampledQuantile):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, n_samples=10, **kwargs)


class Window100(kaczmarz.WindowedQuantile):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, window_size=100, **kwargs)


quantile_strategies.append(Sample100)
quantile_strategies.append(Window100)


@pytest.mark.parametrize("A,x_exact", get_systems())
@pytest.mark.parametrize("QuantileStrategy", quantile_strategies)
def test_corrupted_systems(A, x_exact, QuantileStrategy):
    maxiter = 100  # 100 iterations should be plenty for nice systems.
    # TODO: Add a proper convergence criterion for corrupted systems.
    b = A @ x_exact
    b[0] += 100
    x_approx = QuantileStrategy.solve(A, b, maxiter=maxiter, quantile=0.7)

    residual = b - A @ x_approx

    tol = 1e-5

    assert residual[0] > 99
    assert np.linalg.norm(residual[1:]) < tol



import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as spla


def compute_row_norms(A):
    """Compute the norm of each row of a matrix.

    Parameters
    ----------
    A : (m, n) spmatrx or array_like

    Returns
    -------
    row_norms : (m,) array
    """
    if sp.issparse(A):
        return spla.norm(A, axis=1)

    return np.linalg.norm(A, axis=1)


def normalize_matrix(A, row_norms):
    """Normalize a matrix to have rows with norm 1.

    Parameters
    ----------
    A : (m, n) spmatrx or array
    row_norms : (m,) array

    Returns
    -------
    A_normalized : (m, n) spmatrx or array
    """

    # Be careful! Do not try ``A / row_norms[:, None]`` with a sparse matrix!
    # You will end up with a np.matrix rather than a sparse matrix.

    normalization_matrix = sp.diags(1 / row_norms)
    return normalization_matrix @ A


def normalize_system(A, b):
    """Scale the system ``A @ x = b`` so that the rows of ``A`` have norm 1.

    Parameters
    ----------
    A : (m, n) spmatrix or array_like
    b : (m,) or (m, 1) array_like

    Returns
    -------
    A_normalized : (m, n) array or spmatrx
        Copy of ``A`` with rows scaled to have norm ``1``.
    b_normalized : (m,) or (m, 1) array
        Copy of ``b`` with entries divided by the row norms of ``A``.
    """
    if not sp.issparse(A):
        A = np.array(A)

    row_norms = compute_row_norms(A)
    A = normalize_matrix(A, row_norms=row_norms)
    b = np.array(b).ravel() / row_norms

    return A, b, row_norms
